@@ -1,6 +1,6 @@
 //
 //  Copyright (C) 2001-2004 HorizonLive.com, Inc.  All Rights Reserved.
-//  Copyright (C) 2001,2002 Constantin Kaplinsky.  All Rights Reserved.
+//  Copyright (C) 2001-2006 Constantin Kaplinsky.  All Rights Reserved.
 //  Copyright (C) 2000 Tridia Corporation.  All Rights Reserved.
 //  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 //
@@ -34,7 +34,8 @@ class RfbProto {
 
   final static String
     versionMsg_3_3 = "RFB 003.003\n",
-    versionMsg_3_7 = "RFB 003.007\n";
+    versionMsg_3_7 = "RFB 003.007\n",
+    versionMsg_3_8 = "RFB 003.008\n";
 
   // Vendor signatures: standard VNC/RealVNC, TridiaVNC, and TightVNC
   final static String
@@ -71,14 +72,20 @@ class RfbProto {
     VncAuthFailed  = 1,
     VncAuthTooMany = 2;
 
-  // Server-to-client messages
+  // Standard server-to-client messages
   final static int
     FramebufferUpdate   = 0,
     SetColourMapEntries = 1,
     Bell                = 2,
     ServerCutText       = 3;
 
-  // Client-to-server messages
+  // Non-standard server-to-client messages
+  final static int
+    EndOfContinuousUpdates = 150;
+  final static String
+    SigEndOfContinuousUpdates = "CUS_EOCU";
+
+  // Standard client-to-server messages
   final static int
     SetPixelFormat           = 0,
     FixColourMapEntries      = 1,
@@ -88,6 +95,12 @@ class RfbProto {
     PointerEvent             = 5,
     ClientCutText            = 6;
 
+  // Non-standard client-to-server messages
+  final static int
+    EnableContinuousUpdates = 150;
+  final static String
+    SigEnableContinuousUpdates = "CUC_ENCU";
+
   // Supported encodings and pseudo-encodings
   final static int
     EncodingRaw            = 0,
@@ -97,6 +110,7 @@ class RfbProto {
     EncodingHextile        = 5,
     EncodingZlib           = 6,
     EncodingTight          = 7,
+    EncodingZRLE           = 16,
     EncodingCompressLevel0 = 0xFFFFFF00,
     EncodingQualityLevel0  = 0xFFFFFFE0,
     EncodingXCursor        = 0xFFFFFF10,
@@ -112,6 +126,7 @@ class RfbProto {
     SigEncodingHextile        = "HEXTILE_",
     SigEncodingZlib           = "ZLIB____",
     SigEncodingTight          = "TIGHT___",
+    SigEncodingZRLE           = "ZRLE____",
     SigEncodingCompressLevel0 = "COMPRLVL",
     SigEncodingQualityLevel0  = "JPEGQLVL",
     SigEncodingXCursor        = "X11CURSR",
@@ -120,7 +135,7 @@ class RfbProto {
     SigEncodingLastRect       = "LASTRECT",
     SigEncodingNewFBSize      = "NEWFBSIZ";
 
-  final static int MaxNormalEncoding = 7;
+  final static int MaxNormalEncoding = 255;
 
   // Contstants used in the Hextile decoder
   final static int
@@ -145,12 +160,18 @@ class RfbProto {
   String host;
   int port;
   Socket sock;
-  DataInputStream is;
   OutputStream os;
   SessionRecorder rec;
   boolean inNormalProtocol = false;
   VncViewer viewer;
 
+  // Input stream is declared private to make sure it can be accessed
+  // only via RfbProto methods. We have to do this because we want to
+  // count how many bytes were read.
+  private DataInputStream is;
+  private long numBytesRead = 0;
+  public long getNumBytesRead() { return numBytesRead; }
+
   // Java on UNIX does not call keyPressed() on some keys, for example
   // swedish keys To prevent our workaround to produce duplicate
   // keypresses on JVMs that actually works, keep track of if
@@ -158,12 +179,12 @@ class RfbProto {
   boolean brokenKeyPressed = false;
 
   // This will be set to true on the first framebuffer update
-  // containing Zlib- or Tight-encoded data.
+  // containing Zlib-, ZRLE- or Tight-encoded data.
   boolean wereZlibUpdates = false;
 
   // This will be set to false if the startSession() was called after
-  // we have received at least one Zlib- or Tight-encoded framebuffer
-  // update.
+  // we have received at least one Zlib-, ZRLE- or Tight-encoded
+  // framebuffer update.
   boolean recordFromBeginning = true;
 
   // This fields are needed to show warnings about inefficiently saved
@@ -282,7 +303,10 @@ class RfbProto {
 
   void writeVersionMsg() throws IOException {
     clientMajor = 3;
-    if (serverMajor > 3 || serverMinor >= 7) {
+    if (serverMajor > 3 || serverMinor >= 8) {
+      clientMinor = 8;
+      os.write(versionMsg_3_8.getBytes());
+    } else if (serverMinor >= 7) {
       clientMinor = 7;
       os.write(versionMsg_3_7.getBytes());
     } else {
@@ -290,6 +314,7 @@ class RfbProto {
       os.write(versionMsg_3_3.getBytes());
     }
     protocolTightVNC = false;
+    initCapabilities();
   }
 
 
@@ -307,7 +332,7 @@ class RfbProto {
   //
 
   int readSecurityType() throws Exception {
-    int secType = is.readInt();
+    int secType = readU32();
 
     switch (secType) {
     case SecTypeInvalid:
@@ -322,14 +347,14 @@ class RfbProto {
   }
 
   //
-  // Select security type from the server's list (protocol version 3.7).
+  // Select security type from the server's list (protocol versions 3.7/3.8).
   //
 
   int selectSecurityType() throws Exception {
     int secType = SecTypeInvalid;
 
     // Read the list of secutiry types.
-    int nSecTypes = is.readUnsignedByte();
+    int nSecTypes = readU8();
     if (nSecTypes == 0) {
       readConnFailedReason();
       return SecTypeInvalid;	// should never be executed
@@ -363,20 +388,81 @@ class RfbProto {
     return secType;
   }
 
+  //
+  // Perform "no authentication".
+  //
+
+  void authenticateNone() throws Exception {
+    if (clientMinor >= 8)
+      readSecurityResult("No authentication");
+  }
+
+  //
+  // Perform standard VNC Authentication.
+  //
+
+  void authenticateVNC(String pw) throws Exception {
+    byte[] challenge = new byte[16];
+    readFully(challenge);
+
+    if (pw.length() > 8)
+      pw = pw.substring(0, 8);	// Truncate to 8 chars
+
+    // Truncate password on the first zero byte.
+    int firstZero = pw.indexOf(0);
+    if (firstZero != -1)
+      pw = pw.substring(0, firstZero);
+
+    byte[] key = {0, 0, 0, 0, 0, 0, 0, 0};
+    System.arraycopy(pw.getBytes(), 0, key, 0, pw.length());
+
+    DesCipher des = new DesCipher(key);
+
+    des.encrypt(challenge, 0, challenge, 0);
+    des.encrypt(challenge, 8, challenge, 8);
+
+    os.write(challenge);
+
+    readSecurityResult("VNC authentication");
+  }
+
+  //
+  // Read security result.
+  // Throws an exception on authentication failure.
+  //
+
+  void readSecurityResult(String authType) throws Exception {
+    int securityResult = readU32();
+
+    switch (securityResult) {
+    case VncAuthOK:
+      System.out.println(authType + ": success");
+      break;
+    case VncAuthFailed:
+      if (clientMinor >= 8)
+        readConnFailedReason();
+      throw new Exception(authType + ": failed");
+    case VncAuthTooMany:
+      throw new Exception(authType + ": failed, too many tries");
+    default:
+      throw new Exception(authType + ": unknown result " + securityResult);
+    }
+  }
+
   //
   // Read the string describing the reason for a connection failure,
   // and throw an exception.
   //
 
   void readConnFailedReason() throws Exception {
-    int reasonLen = is.readInt();
+    int reasonLen = readU32();
     byte[] reason = new byte[reasonLen];
     readFully(reason);
     throw new Exception(new String(reason));
   }
 
   //
-  // Initialize capability lists (protocol 3.7t).
+  // Initialize capability lists (TightVNC protocol extensions).
   //
 
   void initCapabilities() {
@@ -391,8 +477,12 @@ class RfbProto {
 		 "No authentication");
     authCaps.add(AuthVNC, StandardVendor, SigAuthVNC,
 		 "Standard VNC password authentication");
-    authCaps.add(AuthUnixLogin, TightVncVendor, SigAuthUnixLogin,
-		 "Login-style Unix authentication");
+
+    // Supported non-standard server-to-client messages
+    // [NONE]
+
+    // Supported non-standard client-to-server messages
+    // [NONE]
 
     // Supported encoding types
     encodingCaps.add(EncodingCopyRect, StandardVendor,
@@ -403,6 +493,8 @@ class RfbProto {
 		     SigEncodingCoRRE, "Standard CoRRE encoding");
     encodingCaps.add(EncodingHextile, StandardVendor,
 		     SigEncodingHextile, "Standard Hextile encoding");
+    encodingCaps.add(EncodingZRLE, StandardVendor,
+		     SigEncodingZRLE, "Standard ZRLE encoding");
     encodingCaps.add(EncodingZlib, TridiaVncVendor,
 		     SigEncodingZlib, "Zlib encoding");
     encodingCaps.add(EncodingTight, TightVncVendor,
@@ -426,11 +518,11 @@ class RfbProto {
   }
 
   //
-  // Setup tunneling (protocol version 3.7t)
+  // Setup tunneling (TightVNC protocol extensions)
   //
 
   void setupTunneling() throws IOException {
-    int nTunnelTypes = is.readInt();
+    int nTunnelTypes = readU32();
     if (nTunnelTypes != 0) {
       readCapabilityList(tunnelCaps, nTunnelTypes);
 
@@ -440,20 +532,18 @@ class RfbProto {
   }
 
   //
-  // Negotiate authentication scheme (protocol version 3.7t)
+  // Negotiate authentication scheme (TightVNC protocol extensions)
   //
 
   int negotiateAuthenticationTight() throws Exception {
-    int nAuthTypes = is.readInt();
+    int nAuthTypes = readU32();
     if (nAuthTypes == 0)
       return AuthNone;
 
     readCapabilityList(authCaps, nAuthTypes);
     for (int i = 0; i < authCaps.numEnabled(); i++) {
       int authType = authCaps.getByOrder(i);
-      if (authType == AuthNone ||
-	  authType == AuthVNC  ||
-	  authType == AuthUnixLogin) {
+      if (authType == AuthNone || authType == AuthVNC) {
 	writeInt(authType);
 	return authType;
       }
@@ -462,7 +552,7 @@ class RfbProto {
   }
 
   //
-  // Read a capability list (protocol version 3.7t)
+  // Read a capability list (TightVNC protocol extensions)
   //
 
   void readCapabilityList(CapsContainer caps, int count) throws IOException {
@@ -470,7 +560,7 @@ class RfbProto {
     byte[] vendor = new byte[4];
     byte[] name = new byte[8];
     for (int i = 0; i < count; i++) {
-      code = is.readInt();
+      code = readU32();
       readFully(vendor);
       readFully(name);
       caps.enable(new CapabilityInfo(code, vendor, name));
@@ -515,31 +605,31 @@ class RfbProto {
   int redMax, greenMax, blueMax, redShift, greenShift, blueShift;
 
   void readServerInit() throws IOException {
-    framebufferWidth = is.readUnsignedShort();
-    framebufferHeight = is.readUnsignedShort();
-    bitsPerPixel = is.readUnsignedByte();
-    depth = is.readUnsignedByte();
-    bigEndian = (is.readUnsignedByte() != 0);
-    trueColour = (is.readUnsignedByte() != 0);
-    redMax = is.readUnsignedShort();
-    greenMax = is.readUnsignedShort();
-    blueMax = is.readUnsignedShort();
-    redShift = is.readUnsignedByte();
-    greenShift = is.readUnsignedByte();
-    blueShift = is.readUnsignedByte();
+    framebufferWidth = readU16();
+    framebufferHeight = readU16();
+    bitsPerPixel = readU8();
+    depth = readU8();
+    bigEndian = (readU8() != 0);
+    trueColour = (readU8() != 0);
+    redMax = readU16();
+    greenMax = readU16();
+    blueMax = readU16();
+    redShift = readU8();
+    greenShift = readU8();
+    blueShift = readU8();
     byte[] pad = new byte[3];
     readFully(pad);
-    int nameLength = is.readInt();
+    int nameLength = readU32();
     byte[] name = new byte[nameLength];
     readFully(name);
     desktopName = new String(name);
 
-    // Read interaction capabilities (protocol 3.7t)
+    // Read interaction capabilities (TightVNC protocol extensions)
     if (protocolTightVNC) {
-      int nServerMessageTypes = is.readUnsignedShort();
-      int nClientMessageTypes = is.readUnsignedShort();
-      int nEncodingTypes = is.readUnsignedShort();
-      is.readUnsignedShort();
+      int nServerMessageTypes = readU16();
+      int nClientMessageTypes = readU16();
+      int nEncodingTypes = readU16();
+      readU16();
       readCapabilityList(serverMsgCaps, nServerMessageTypes);
       readCapabilityList(clientMsgCaps, nClientMessageTypes);
       readCapabilityList(encodingCaps, nEncodingTypes);
@@ -570,6 +660,11 @@ class RfbProto {
     rec.write(desktopName.getBytes());
     numUpdatesInSession = 0;
 
+    // FIXME: If there were e.g. ZRLE updates only, that should not
+    //        affect recording of Zlib and Tight updates. So, actually
+    //        we should maintain separate flags for Zlib, ZRLE and
+    //        Tight, instead of one ``wereZlibUpdates'' variable.
+    //
     if (wereZlibUpdates)
       recordFromBeginning = false;
 
@@ -604,7 +699,7 @@ class RfbProto {
   //
 
   int readServerMessageType() throws IOException {
-    int msgType = is.readUnsignedByte();
+    int msgType = readU8();
 
     // If the session is being recorded:
     if (rec != null) {
@@ -626,8 +721,8 @@ class RfbProto {
   int updateNRects;
 
   void readFramebufferUpdate() throws IOException {
-    is.readByte();
-    updateNRects = is.readUnsignedShort();
+    skipBytes(1);
+    updateNRects = readU16();
 
     // If the session is being recorded:
     if (rec != null) {
@@ -644,13 +739,14 @@ class RfbProto {
   int updateRectX, updateRectY, updateRectW, updateRectH, updateRectEncoding;
 
   void readFramebufferUpdateRectHdr() throws Exception {
-    updateRectX = is.readUnsignedShort();
-    updateRectY = is.readUnsignedShort();
-    updateRectW = is.readUnsignedShort();
-    updateRectH = is.readUnsignedShort();
-    updateRectEncoding = is.readInt();
+    updateRectX = readU16();
+    updateRectY = readU16();
+    updateRectW = readU16();
+    updateRectH = readU16();
+    updateRectEncoding = readU32();
 
     if (updateRectEncoding == EncodingZlib ||
+        updateRectEncoding == EncodingZRLE ||
 	updateRectEncoding == EncodingTight)
       wereZlibUpdates = true;
 
@@ -698,8 +794,8 @@ class RfbProto {
   int copyRectSrcX, copyRectSrcY;
 
   void readCopyRect() throws IOException {
-    copyRectSrcX = is.readUnsignedShort();
-    copyRectSrcY = is.readUnsignedShort();
+    copyRectSrcX = readU16();
+    copyRectSrcY = readU16();
 
     // If the session is being recorded:
     if (rec != null) {
@@ -714,9 +810,8 @@ class RfbProto {
   //
 
   String readServerCutText() throws IOException {
-    byte[] pad = new byte[3];
-    readFully(pad);
-    int len = is.readInt();
+    skipBytes(3);
+    int len = readU32();
     byte[] text = new byte[len];
     readFully(text);
     return new String(text);
@@ -732,15 +827,15 @@ class RfbProto {
 
   int readCompactLen() throws IOException {
     int[] portion = new int[3];
-    portion[0] = is.readUnsignedByte();
+    portion[0] = readU8();
     int byteCount = 1;
     int len = portion[0] & 0x7F;
     if ((portion[0] & 0x80) != 0) {
-      portion[1] = is.readUnsignedByte();
+      portion[1] = readU8();
       byteCount++;
       len |= (portion[1] & 0x7F) << 7;
       if ((portion[1] & 0x80) != 0) {
-	portion[2] = is.readUnsignedByte();
+	portion[2] = readU8();
 	byteCount++;
 	len |= (portion[2] & 0xFF) << 14;
       }
@@ -1145,7 +1240,6 @@ class RfbProto {
     oldModifiers = newModifiers;
   }
 
-
   //
   // Compress and write the data into the recorded session file. This
   // method assumes the recording is on (rec != null).
@@ -1212,6 +1306,13 @@ class RfbProto {
     return timeWaitedIn100us;
   }
 
+  //
+  // Methods for reading data via our DataInputStream member variable (is).
+  //
+  // In addition to reading data, the readFully() methods updates variables
+  // used to estimate data throughput.
+  //
+
   public void readFully(byte b[]) throws IOException {
     readFully(b, 0, b.length);
   }
@@ -1236,6 +1337,38 @@ class RfbProto {
       timeWaitedIn100us += newTimeWaited;
       timedKbits += newKbits;
     }
+
+    numBytesRead += len;
   }
 
+  final int available() throws IOException {
+    return is.available();
+  }
+
+  // FIXME: DataInputStream::skipBytes() is not guaranteed to skip
+  //        exactly n bytes. Probably we don't want to use this method.
+  final int skipBytes(int n) throws IOException {
+    int r = is.skipBytes(n);
+    numBytesRead += r;
+    return r;
+  }
+
+  final int readU8() throws IOException {
+    int r = is.readUnsignedByte();
+    numBytesRead++;
+    return r;
+  }
+
+  final int readU16() throws IOException {
+    int r = is.readUnsignedShort();
+    numBytesRead += 2;
+    return r;
+  }
+
+  final int readU32() throws IOException {
+    int r = is.readInt();
+    numBytesRead += 4;
+    return r;
+  }
 }
+
