@@ -42,6 +42,10 @@ class VncCanvas extends Canvas
   Color[] colors;
   int bytesPixel;
 
+  int maxWidth = 0, maxHeight = 0;
+  int scalingFactor;
+  int scaledWidth, scaledHeight;
+
   Image memImage;
   Graphics memGraphics;
 
@@ -50,6 +54,28 @@ class VncCanvas extends Canvas
   byte[] pixels8;
   int[] pixels24;
 
+  // Update statistics.
+  long statStartTime;           // time on first framebufferUpdateRequest
+  int statNumUpdates;           // counter for FramebufferUpdate messages
+  int statNumTotalRects;        // rectangles in FramebufferUpdate messages
+  int statNumPixelRects;        // the same, but excluding pseudo-rectangles
+  int statNumRectsTight;        // Tight-encoded rectangles (including JPEG)
+  int statNumRectsTightJPEG;    // JPEG-compressed Tight-encoded rectangles
+  int statNumRectsZRLE;         // ZRLE-encoded rectangles
+  int statNumRectsHextile;      // Hextile-encoded rectangles
+  int statNumRectsRaw;          // Raw-encoded rectangles
+  int statNumRectsCopy;         // CopyRect rectangles
+  int statNumBytesEncoded;      // number of bytes in updates, as received
+  int statNumBytesDecoded;      // number of bytes, as if Raw encoding was used
+
+  // ZRLE encoder's data.
+  byte[] zrleBuf;
+  int zrleBufLen = 0;
+  byte[] zrleTilePixels8;
+  int[] zrleTilePixels24;
+  ZlibInStream zrleInStream;
+  boolean zrleRecWarningShown = false;
+
   // Zlib encoder's data.
   byte[] zlibBuf;
   int zlibBufLen = 0;
@@ -69,12 +95,18 @@ class VncCanvas extends Canvas
   boolean inputEnabled;
 
   //
-  // The constructor.
+  // The constructors.
   //
 
-  VncCanvas(VncViewer v) throws IOException {
+  public VncCanvas(VncViewer v, int maxWidth_, int maxHeight_)
+    throws IOException {
+
     viewer = v;
+    maxWidth = maxWidth_;
+    maxHeight = maxHeight_;
+
     rfb = viewer.rfb;
+    scalingFactor = viewer.options.scalingFactor;
 
     tightInflaters = new Inflater[4];
 
@@ -96,20 +128,24 @@ class VncCanvas extends Canvas
     addKeyListener(this);
   }
 
+  public VncCanvas(VncViewer v) throws IOException {
+    this(v, 0, 0);
+  }
+
   //
   // Callback methods to determine geometry of our Component.
   //
 
   public Dimension getPreferredSize() {
-    return new Dimension(rfb.framebufferWidth, rfb.framebufferHeight);
+    return new Dimension(scaledWidth, scaledHeight);
   }
 
   public Dimension getMinimumSize() {
-    return new Dimension(rfb.framebufferWidth, rfb.framebufferHeight);
+    return new Dimension(scaledWidth, scaledHeight);
   }
 
   public Dimension getMaximumSize() {
-    return new Dimension(rfb.framebufferWidth, rfb.framebufferHeight);
+    return new Dimension(scaledWidth, scaledHeight);
   }
 
   //
@@ -122,7 +158,11 @@ class VncCanvas extends Canvas
 
   public void paint(Graphics g) {
     synchronized(memImage) {
-      g.drawImage(memImage, 0, 0, null);
+      if (rfb.framebufferWidth == scaledWidth) {
+        g.drawImage(memImage, 0, 0, null);
+      } else {
+        paintScaledFrameBuffer(g);
+      }
     }
     if (showSoftCursor) {
       int x0 = cursorX - hotX, y0 = cursorY - hotY;
@@ -133,6 +173,10 @@ class VncCanvas extends Canvas
     }
   }
 
+  public void paintScaledFrameBuffer(Graphics g) {
+    g.drawImage(memImage, 0, 0, scaledWidth, scaledHeight, null);
+  }
+
   //
   // Override the ImageObserver interface method to handle drawing of
   // JPEG-encoded data.
@@ -201,6 +245,20 @@ class VncCanvas extends Canvas
     int fbWidth = rfb.framebufferWidth;
     int fbHeight = rfb.framebufferHeight;
 
+    // Calculate scaling factor for auto scaling.
+    if (maxWidth > 0 && maxHeight > 0) {
+      int f1 = maxWidth * 100 / fbWidth;
+      int f2 = maxHeight * 100 / fbHeight;
+      scalingFactor = Math.min(f1, f2);
+      if (scalingFactor > 100)
+	scalingFactor = 100;
+      System.out.println("Scaling desktop at " + scalingFactor + "%");
+    }
+
+    // Update scaled framebuffer geometry.
+    scaledWidth = (fbWidth * scalingFactor + 50) / 100;
+    scaledHeight = (fbHeight * scalingFactor + 50) / 100;
+
     // Create new off-screen image either if it does not exist, or if
     // its geometry should be changed. It's not necessary to replace
     // existing image if only pixel format should be changed.
@@ -218,17 +276,27 @@ class VncCanvas extends Canvas
     // Images with raw pixels should be re-allocated on every change
     // of geometry or pixel format.
     if (bytesPixel == 1) {
+
       pixels24 = null;
       pixels8 = new byte[fbWidth * fbHeight];
 
       pixelsSource =
 	new MemoryImageSource(fbWidth, fbHeight, cm8, pixels8, 0, fbWidth);
+
+      zrleTilePixels24 = null;
+      zrleTilePixels8 = new byte[64 * 64];
+
     } else {
+
       pixels8 = null;
       pixels24 = new int[fbWidth * fbHeight];
 
       pixelsSource =
 	new MemoryImageSource(fbWidth, fbHeight, cm24, pixels24, 0, fbWidth);
+
+      zrleTilePixels8 = null;
+      zrleTilePixels24 = new int[64 * 64];
+
     }
     pixelsSource.setAnimated(true);
     rawPixelsImage = Toolkit.getDefaultToolkit().createImage(pixelsSource);
@@ -238,20 +306,20 @@ class VncCanvas extends Canvas
       if (viewer.desktopScrollPane != null)
 	resizeDesktopFrame();
     } else {
-      setSize(fbWidth, fbHeight);
+      setSize(scaledWidth, scaledHeight);
     }
     viewer.moveFocusToDesktop();
   }
 
   void resizeDesktopFrame() {
-    setSize(rfb.framebufferWidth, rfb.framebufferHeight);
+    setSize(scaledWidth, scaledHeight);
 
     // FIXME: Find a better way to determine correct size of a
     // ScrollPane.  -- const
     Insets insets = viewer.desktopScrollPane.getInsets();
-    viewer.desktopScrollPane.setSize(rfb.framebufferWidth +
+    viewer.desktopScrollPane.setSize(scaledWidth +
 				     2 * Math.min(insets.left, insets.right),
-				     rfb.framebufferHeight +
+				     scaledHeight +
 				     2 * Math.min(insets.top, insets.bottom));
 
     viewer.vncFrame.pack();
@@ -301,6 +369,9 @@ class VncCanvas extends Canvas
     rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
 				      rfb.framebufferHeight, false);
 
+    resetStats();
+    boolean statsRestarted = false;
+
     //
     // main dispatch loop
     //
@@ -313,12 +384,25 @@ class VncCanvas extends Canvas
       // Process the message depending on its type.
       switch (msgType) {
       case RfbProto.FramebufferUpdate:
+
+        if (statNumUpdates == viewer.debugStatsExcludeUpdates &&
+            !statsRestarted) {
+          resetStats();
+          statsRestarted = true;
+        } else if (statNumUpdates == viewer.debugStatsMeasureUpdates &&
+                   statsRestarted) {
+          viewer.disconnect();
+        }
+
 	rfb.readFramebufferUpdate();
+	statNumUpdates++;
 
 	boolean cursorPosReceived = false;
 
 	for (int i = 0; i < rfb.updateNRects; i++) {
+
 	  rfb.readFramebufferUpdateRectHdr();
+	  statNumTotalRects++;
 	  int rx = rfb.updateRectX, ry = rfb.updateRectY;
 	  int rw = rfb.updateRectW, rh = rfb.updateRectH;
 
@@ -343,13 +427,17 @@ class VncCanvas extends Canvas
 	    continue;
 	  }
 
+          long numBytesReadBefore = rfb.getNumBytesRead();
+
           rfb.startTiming();
 
 	  switch (rfb.updateRectEncoding) {
 	  case RfbProto.EncodingRaw:
+	    statNumRectsRaw++;
 	    handleRawRect(rx, ry, rw, rh);
 	    break;
 	  case RfbProto.EncodingCopyRect:
+	    statNumRectsCopy++;
 	    handleCopyRect(rx, ry, rw, rh);
 	    break;
 	  case RfbProto.EncodingRRE:
@@ -359,12 +447,18 @@ class VncCanvas extends Canvas
 	    handleCoRRERect(rx, ry, rw, rh);
 	    break;
 	  case RfbProto.EncodingHextile:
+	    statNumRectsHextile++;
 	    handleHextileRect(rx, ry, rw, rh);
 	    break;
+	  case RfbProto.EncodingZRLE:
+	    statNumRectsZRLE++;
+	    handleZRLERect(rx, ry, rw, rh);
+	    break;
 	  case RfbProto.EncodingZlib:
             handleZlibRect(rx, ry, rw, rh);
 	    break;
 	  case RfbProto.EncodingTight:
+	    statNumRectsTight++;
 	    handleTightRect(rx, ry, rw, rh);
 	    break;
 	  default:
@@ -373,6 +467,11 @@ class VncCanvas extends Canvas
 	  }
 
           rfb.stopTiming();
+
+          statNumPixelRects++;
+          statNumBytesDecoded += rw * rh * bytesPixel;
+          statNumBytesEncoded +=
+            (int)(rfb.getNumBytesRead() - numBytesReadBefore);
 	}
 
 	boolean fullUpdateNeeded = false;
@@ -387,7 +486,7 @@ class VncCanvas extends Canvas
 	// if there is some data to receive, or if the last update
 	// included a PointerPos message.
 	if (viewer.deferUpdateRequests > 0 &&
-	    rfb.is.available() == 0 && !cursorPosReceived) {
+	    rfb.available() == 0 && !cursorPosReceived) {
 	  synchronized(rfb) {
 	    try {
 	      rfb.wait(viewer.deferUpdateRequests);
@@ -396,19 +495,20 @@ class VncCanvas extends Canvas
 	  }
 	}
 
-	// Before requesting framebuffer update, check if the pixel
-	// format should be changed. If it should, request full update
-	// instead of an incremental one.
-	if (viewer.options.eightBitColors != (bytesPixel == 1)) {
-	  setPixelFormat();
-	  fullUpdateNeeded = true;
-	}
-
         viewer.autoSelectEncodings();
 
-	rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
-					  rfb.framebufferHeight,
-					  !fullUpdateNeeded);
+	// Before requesting framebuffer update, check if the pixel
+	// format should be changed.
+	if (viewer.options.eightBitColors != (bytesPixel == 1)) {
+          // Pixel format should be changed.
+          setPixelFormat();
+          fullUpdateNeeded = true;
+	}
+
+        // Request framebuffer update if needed.
+        int w = rfb.framebufferWidth;
+        int h = rfb.framebufferHeight;
+        rfb.writeFramebufferUpdateRequest(0, 0, w, h, !fullUpdateNeeded);
 
 	break;
 
@@ -492,7 +592,7 @@ class VncCanvas extends Canvas
 
   void handleRRERect(int x, int y, int w, int h) throws IOException {
 
-    int nSubrects = rfb.is.readInt();
+    int nSubrects = rfb.readU32();
 
     byte[] bg_buf = new byte[bytesPixel];
     rfb.readFully(bg_buf);
@@ -543,7 +643,7 @@ class VncCanvas extends Canvas
   //
 
   void handleCoRRERect(int x, int y, int w, int h) throws IOException {
-    int nSubrects = rfb.is.readInt();
+    int nSubrects = rfb.readU32();
 
     byte[] bg_buf = new byte[bytesPixel];
     rfb.readFully(bg_buf);
@@ -624,7 +724,7 @@ class VncCanvas extends Canvas
   void handleHextileSubrect(int tx, int ty, int tw, int th)
     throws IOException {
 
-    int subencoding = rfb.is.readUnsignedByte();
+    int subencoding = rfb.readU8();
     if (rfb.rec != null) {
       rfb.rec.writeByte(subencoding);
     }
@@ -668,7 +768,7 @@ class VncCanvas extends Canvas
     if ((subencoding & rfb.HextileAnySubrects) == 0)
       return;
 
-    int nSubrects = rfb.is.readUnsignedByte();
+    int nSubrects = rfb.readU8();
     int bufsize = nSubrects * 2;
     if ((subencoding & rfb.HextileSubrectsColoured) != 0) {
       bufsize += nSubrects * bytesPixel;
@@ -732,13 +832,244 @@ class VncCanvas extends Canvas
     }
   }
 
+  //
+  // Handle a ZRLE-encoded rectangle.
+  //
+  // FIXME: Currently, session recording is not fully supported for ZRLE.
+  //
+
+  void handleZRLERect(int x, int y, int w, int h) throws Exception {
+
+    if (zrleInStream == null)
+      zrleInStream = new ZlibInStream();
+
+    int nBytes = rfb.readU32();
+    if (nBytes > 64 * 1024 * 1024)
+      throw new Exception("ZRLE decoder: illegal compressed data size");
+
+    if (zrleBuf == null || zrleBufLen < nBytes) {
+      zrleBufLen = nBytes + 4096;
+      zrleBuf = new byte[zrleBufLen];
+    }
+
+    // FIXME: Do not wait for all the data before decompression.
+    rfb.readFully(zrleBuf, 0, nBytes);
+
+    if (rfb.rec != null) {
+      if (rfb.recordFromBeginning) {
+        rfb.rec.writeIntBE(nBytes);
+        rfb.rec.write(zrleBuf, 0, nBytes);
+      } else if (!zrleRecWarningShown) {
+        System.out.println("Warning: ZRLE session can be recorded" +
+                           " only from the beginning");
+        System.out.println("Warning: Recorded file may be corrupted");
+        zrleRecWarningShown = true;
+      }
+    }
+
+    zrleInStream.setUnderlying(new MemInStream(zrleBuf, 0, nBytes), nBytes);
+
+    for (int ty = y; ty < y+h; ty += 64) {
+
+      int th = Math.min(y+h-ty, 64);
+
+      for (int tx = x; tx < x+w; tx += 64) {
+
+        int tw = Math.min(x+w-tx, 64);
+
+        int mode = zrleInStream.readU8();
+        boolean rle = (mode & 128) != 0;
+        int palSize = mode & 127;
+        int[] palette = new int[128];
+
+        readZrlePalette(palette, palSize);
+
+        if (palSize == 1) {
+          int pix = palette[0];
+          Color c = (bytesPixel == 1) ?
+            colors[pix] : new Color(0xFF000000 | pix);
+          memGraphics.setColor(c);
+          memGraphics.fillRect(tx, ty, tw, th);
+          continue;
+        }
+
+        if (!rle) {
+          if (palSize == 0) {
+            readZrleRawPixels(tw, th);
+          } else {
+            readZrlePackedPixels(tw, th, palette, palSize);
+          }
+        } else {
+          if (palSize == 0) {
+            readZrlePlainRLEPixels(tw, th);
+          } else {
+            readZrlePackedRLEPixels(tw, th, palette);
+          }
+        }
+        handleUpdatedZrleTile(tx, ty, tw, th);
+      }
+    }
+
+    zrleInStream.reset();
+
+    scheduleRepaint(x, y, w, h);
+  }
+
+  int readPixel(InStream is) throws Exception {
+    int pix;
+    if (bytesPixel == 1) {
+      pix = is.readU8();
+    } else {
+      int p1 = is.readU8();
+      int p2 = is.readU8();
+      int p3 = is.readU8();
+      pix = (p3 & 0xFF) << 16 | (p2 & 0xFF) << 8 | (p1 & 0xFF);
+    }
+    return pix;
+  }
+
+  void readPixels(InStream is, int[] dst, int count) throws Exception {
+    int pix;
+    if (bytesPixel == 1) {
+      byte[] buf = new byte[count];
+      is.readBytes(buf, 0, count);
+      for (int i = 0; i < count; i++) {
+        dst[i] = (int)buf[i] & 0xFF;
+      }
+    } else {
+      byte[] buf = new byte[count * 3];
+      is.readBytes(buf, 0, count * 3);
+      for (int i = 0; i < count; i++) {
+        dst[i] = ((buf[i*3+2] & 0xFF) << 16 |
+                  (buf[i*3+1] & 0xFF) << 8 |
+                  (buf[i*3] & 0xFF));
+      }
+    }
+  }
+
+  void readZrlePalette(int[] palette, int palSize) throws Exception {
+    readPixels(zrleInStream, palette, palSize);
+  }
+
+  void readZrleRawPixels(int tw, int th) throws Exception {
+    if (bytesPixel == 1) {
+      zrleInStream.readBytes(zrleTilePixels8, 0, tw * th);
+    } else {
+      readPixels(zrleInStream, zrleTilePixels24, tw * th); ///
+    }
+  }
+
+  void readZrlePackedPixels(int tw, int th, int[] palette, int palSize)
+    throws Exception {
+
+    int bppp = ((palSize > 16) ? 8 :
+                ((palSize > 4) ? 4 : ((palSize > 2) ? 2 : 1)));
+    int ptr = 0;
+
+    for (int i = 0; i < th; i++) {
+      int eol = ptr + tw;
+      int b = 0;
+      int nbits = 0;
+
+      while (ptr < eol) {
+        if (nbits == 0) {
+          b = zrleInStream.readU8();
+          nbits = 8;
+        }
+        nbits -= bppp;
+        int index = (b >> nbits) & ((1 << bppp) - 1) & 127;
+        if (bytesPixel == 1) {
+          zrleTilePixels8[ptr++] = (byte)palette[index];
+        } else {
+          zrleTilePixels24[ptr++] = palette[index];
+        }
+      }
+    }
+  }
+
+  void readZrlePlainRLEPixels(int tw, int th) throws Exception {
+    int ptr = 0;
+    int end = ptr + tw * th;
+    while (ptr < end) {
+      int pix = readPixel(zrleInStream);
+      int len = 1;
+      int b;
+      do {
+        b = zrleInStream.readU8();
+        len += b;
+      } while (b == 255);
+
+      if (!(len <= end - ptr))
+        throw new Exception("ZRLE decoder: assertion failed" +
+                            " (len <= end-ptr)");
+
+      if (bytesPixel == 1) {
+        while (len-- > 0) zrleTilePixels8[ptr++] = (byte)pix;
+      } else {
+        while (len-- > 0) zrleTilePixels24[ptr++] = pix;
+      }
+    }
+  }
+
+  void readZrlePackedRLEPixels(int tw, int th, int[] palette)
+    throws Exception {
+
+    int ptr = 0;
+    int end = ptr + tw * th;
+    while (ptr < end) {
+      int index = zrleInStream.readU8();
+      int len = 1;
+      if ((index & 128) != 0) {
+        int b;
+        do {
+          b = zrleInStream.readU8();
+          len += b;
+        } while (b == 255);
+        
+        if (!(len <= end - ptr))
+          throw new Exception("ZRLE decoder: assertion failed" +
+                              " (len <= end - ptr)");
+      }
+
+      index &= 127;
+      int pix = palette[index];
+
+      if (bytesPixel == 1) {
+        while (len-- > 0) zrleTilePixels8[ptr++] = (byte)pix;
+      } else {
+        while (len-- > 0) zrleTilePixels24[ptr++] = pix;
+      }
+    }
+  }
+
+  //
+  // Copy pixels from zrleTilePixels8 or zrleTilePixels24, then update.
+  //
+
+  void handleUpdatedZrleTile(int x, int y, int w, int h) {
+    Object src, dst;
+    if (bytesPixel == 1) {
+      src = zrleTilePixels8; dst = pixels8;
+    } else {
+      src = zrleTilePixels24; dst = pixels24;
+    }
+    int offsetSrc = 0;
+    int offsetDst = (y * rfb.framebufferWidth + x);
+    for (int j = 0; j < h; j++) {
+      System.arraycopy(src, offsetSrc, dst, offsetDst, w);
+      offsetSrc += w;
+      offsetDst += rfb.framebufferWidth;
+    }
+    handleUpdatedPixels(x, y, w, h);
+  }
+
   //
   // Handle a Zlib-encoded rectangle.
   //
 
   void handleZlibRect(int x, int y, int w, int h) throws Exception {
 
-    int nBytes = rfb.is.readInt();
+    int nBytes = rfb.readU32();
 
     if (zlibBuf == null || zlibBufLen < nBytes) {
       zlibBufLen = nBytes * 2;
@@ -790,7 +1121,7 @@ class VncCanvas extends Canvas
 
   void handleTightRect(int x, int y, int w, int h) throws Exception {
 
-    int comp_ctl = rfb.is.readUnsignedByte();
+    int comp_ctl = rfb.readU8();
     if (rfb.rec != null) {
       if (rfb.recordFromBeginning ||
 	  comp_ctl == (rfb.TightFill << 4) ||
@@ -820,7 +1151,7 @@ class VncCanvas extends Canvas
     if (comp_ctl == rfb.TightFill) {
 
       if (bytesPixel == 1) {
-	int idx = rfb.is.readUnsignedByte();
+	int idx = rfb.readU8();
 	memGraphics.setColor(colors[idx]);
 	if (rfb.rec != null) {
 	  rfb.rec.writeByte(idx);
@@ -843,6 +1174,8 @@ class VncCanvas extends Canvas
 
     if (comp_ctl == rfb.TightJpeg) {
 
+      statNumRectsTightJPEG++;
+
       // Read JPEG data.
       byte[] jpegData = new byte[rfb.readCompactLen()];
       rfb.readFully(jpegData);
@@ -883,12 +1216,12 @@ class VncCanvas extends Canvas
     int[] palette24 = new int[256];
     boolean useGradient = false;
     if ((comp_ctl & rfb.TightExplicitFilter) != 0) {
-      int filter_id = rfb.is.readUnsignedByte();
+      int filter_id = rfb.readU8();
       if (rfb.rec != null) {
 	rfb.rec.writeByte(filter_id);
       }
       if (filter_id == rfb.TightFilterPalette) {
-	numColors = rfb.is.readUnsignedByte() + 1;
+	numColors = rfb.readU8() + 1;
 	if (rfb.rec != null) {
 	  rfb.rec.writeByte(numColors - 1);
 	}
@@ -1167,7 +1500,15 @@ class VncCanvas extends Canvas
 
   void scheduleRepaint(int x, int y, int w, int h) {
     // Request repaint, deferred if necessary.
-    repaint(viewer.deferScreenUpdates, x, y, w, h);
+    if (rfb.framebufferWidth == scaledWidth) {
+      repaint(viewer.deferScreenUpdates, x, y, w, h);
+    } else {
+      int sx = x * scalingFactor / 100;
+      int sy = y * scalingFactor / 100;
+      int sw = ((x + w) * scalingFactor + 49) / 100 - sx + 1;
+      int sh = ((y + h) * scalingFactor + 49) / 100 - sy + 1;
+      repaint(viewer.deferScreenUpdates, sx, sy, sw, sh);
+    }
   }
 
   //
@@ -1232,6 +1573,11 @@ class VncCanvas extends Canvas
       if (moved) {
 	softCursorMove(evt.getX(), evt.getY());
       }
+      if (rfb.framebufferWidth != scaledWidth) {
+        int sx = (evt.getX() * 100 + scalingFactor/2) / scalingFactor;
+        int sy = (evt.getY() * 100 + scalingFactor/2) / scalingFactor;
+        evt.translatePoint(sx - evt.getX(), sy - evt.getY());
+      }
       synchronized(rfb) {
 	try {
 	  rfb.writePointerEvent(evt);
@@ -1251,6 +1597,24 @@ class VncCanvas extends Canvas
   public void mouseEntered(MouseEvent evt) {}
   public void mouseExited(MouseEvent evt) {}
 
+  //
+  // Reset update statistics.
+  //
+
+  void resetStats() {
+    statStartTime = System.currentTimeMillis();
+    statNumUpdates = 0;
+    statNumTotalRects = 0;
+    statNumPixelRects = 0;
+    statNumRectsTight = 0;
+    statNumRectsTightJPEG = 0;
+    statNumRectsZRLE = 0;
+    statNumRectsHextile = 0;
+    statNumRectsRaw = 0;
+    statNumRectsCopy = 0;
+    statNumBytesEncoded = 0;
+    statNumBytesDecoded = 0;
+  }
 
   //////////////////////////////////////////////////////////////////
   //
@@ -1288,10 +1652,10 @@ class VncCanvas extends Canvas
       int bytesMaskData = bytesPerRow * height;
 
       if (encodingType == rfb.EncodingXCursor) {
-	rfb.is.skipBytes(6 + bytesMaskData * 2);
+	rfb.skipBytes(6 + bytesMaskData * 2);
       } else {
 	// rfb.EncodingRichCursor
-	rfb.is.skipBytes(width * height + bytesMaskData);
+	rfb.skipBytes(width * height * bytesPixel + bytesMaskData);
       }
       return;
     }
@@ -1393,9 +1757,9 @@ class VncCanvas extends Canvas
 		result = cm8.getRGB(pixBuf[i]);
 	      } else {
 		result = 0xFF000000 |
-		  (pixBuf[i * 4 + 1] & 0xFF) << 16 |
-		  (pixBuf[i * 4 + 2] & 0xFF) << 8 |
-		  (pixBuf[i * 4 + 3] & 0xFF);
+		  (pixBuf[i * 4 + 2] & 0xFF) << 16 |
+		  (pixBuf[i * 4 + 1] & 0xFF) << 8 |
+		  (pixBuf[i * 4] & 0xFF);
 	      }
 	    } else {
 	      result = 0;	// Transparent pixel
@@ -1409,9 +1773,9 @@ class VncCanvas extends Canvas
 	      result = cm8.getRGB(pixBuf[i]);
 	    } else {
 	      result = 0xFF000000 |
-		(pixBuf[i * 4 + 1] & 0xFF) << 16 |
-		(pixBuf[i * 4 + 2] & 0xFF) << 8 |
-		(pixBuf[i * 4 + 3] & 0xFF);
+		(pixBuf[i * 4 + 2] & 0xFF) << 16 |
+		(pixBuf[i * 4 + 1] & 0xFF) << 8 |
+		(pixBuf[i * 4] & 0xFF);
 	    }
 	  } else {
 	    result = 0;		// Transparent pixel
