@@ -62,8 +62,6 @@ public class VncViewer extends java.applet.Applet
   GridBagLayout gridbag;
   ButtonPanel buttonPanel;
   Label connStatusLabel;
-  AuthPanel authenticator;
-  AuthUnixLoginPanel authenticatorUnixLogin;
   VncCanvas vc;
   OptionsFrame options;
   ClipboardFrame clipboard;
@@ -81,12 +79,15 @@ public class VncViewer extends java.applet.Applet
   String socketFactory;
   String host;
   int port;
+  String passwordParam;
   boolean showControls;
   boolean offerRelogin;
   boolean showOfflineDesktop;
   int deferScreenUpdates;
   int deferCursorUpdates;
   int deferUpdateRequests;
+  int debugStatsExcludeUpdates;
+  int debugStatsMeasureUpdates;
 
   // Reference to this applet for inter-applet communication.
   public static java.applet.Applet refApplet;
@@ -115,8 +116,6 @@ public class VncViewer extends java.applet.Applet
 
     options = new OptionsFrame(this);
     clipboard = new ClipboardFrame(this);
-    authenticator = new AuthPanel(this);
-    authenticatorUnixLogin = new AuthUnixLoginPanel();
     if (RecordingFrame.checkSecurity())
       rec = new RecordingFrame(this);
 
@@ -159,7 +158,19 @@ public class VncViewer extends java.applet.Applet
       connectAndAuthenticate();
       doProtocolInitialisation();
 
-      vc = new VncCanvas(this);
+      // FIXME: Use auto-scaling not only in a separate frame.
+      if (options.autoScale && inSeparateFrame) {
+	Dimension screenSize;
+	try {
+	  screenSize = vncContainer.getToolkit().getScreenSize();
+	} catch (Exception e) {
+	  screenSize = new Dimension(0, 0);
+	}
+	createCanvas(screenSize.width - 32, screenSize.height - 32);
+      } else {
+	createCanvas(0, 0);
+      }
+
       gbc.weightx = 1.0;
       gbc.weighty = 1.0;
 
@@ -247,6 +258,33 @@ public class VncViewer extends java.applet.Applet
     
   }
 
+  //
+  // Create a VncCanvas instance.
+  //
+
+  void createCanvas(int maxWidth, int maxHeight) throws IOException {
+    // Determine if Java 2D API is available and use a special
+    // version of VncCanvas if it is present.
+    vc = null;
+    try {
+      // This throws ClassNotFoundException if there is no Java 2D API.
+      Class cl = Class.forName("java.awt.Graphics2D");
+      // If we could load Graphics2D class, then we can use VncCanvas2D.
+      cl = Class.forName("VncCanvas2");
+      Class[] argClasses = { this.getClass(), Integer.TYPE, Integer.TYPE };
+      java.lang.reflect.Constructor cstr = cl.getConstructor(argClasses);
+      Object[] argObjects =
+        { this, new Integer(maxWidth), new Integer(maxHeight) };
+      vc = (VncCanvas)cstr.newInstance(argObjects);
+    } catch (Exception e) {
+      System.out.println("Warning: Java 2D API is not available");
+    }
+
+    // If we failed to create VncCanvas2D, use old VncCanvas.
+    if (vc == null)
+      vc = new VncCanvas(this, maxWidth, maxHeight);
+  }
+
 
   //
   // Process RFB socket messages.
@@ -282,19 +320,6 @@ public class VncViewer extends java.applet.Applet
       validate();
     }
 
-    while (!tryAuthenticate()) {
-      authenticator.retry();
-      authenticatorUnixLogin.retry();
-    }
-  }
-
-
-  //
-  // Try to connect and authenticate.
-  //
-
-  boolean tryAuthenticate() throws Exception
-  {
     showConnectionStatus("Connecting to " + host + ", port " + port + "...");
 
     rfb = new RfbProto(host, port, this);
@@ -309,63 +334,32 @@ public class VncViewer extends java.applet.Applet
 			 rfb.clientMajor + "." + rfb.clientMinor);
 
     int secType = rfb.negotiateSecurity();
-    int authType = 0;
-
-    // FIXME: Map security types to authentication schemes in RfbProto.
-    switch (secType) {
-    case RfbProto.SecTypeNone:
-      authType = RfbProto.AuthNone;
-      break;
-    case RfbProto.SecTypeVncAuth:
-      authType = RfbProto.AuthVNC;
-      break;
-    case RfbProto.SecTypeTight:
+    int authType;
+    if (secType == RfbProto.SecTypeTight) {
       showConnectionStatus("Enabling TightVNC protocol extensions");
-      rfb.initCapabilities();
       rfb.setupTunneling();
       authType = rfb.negotiateAuthenticationTight();
-      break;
-    default:
-      throw new Exception("Unknown security type " + secType);
+    } else {
+      authType = secType;
     }
 
-    boolean success = false;
-
     switch (authType) {
     case RfbProto.AuthNone:
       showConnectionStatus("No authentication needed");
-      success = true;
+      rfb.authenticateNone();
       break;
     case RfbProto.AuthVNC:
       showConnectionStatus("Performing standard VNC authentication");
-      if (authenticator.isInteractionNecessary()) {
-	showAuthPanel(authenticator);
-	authenticator.moveFocusToDefaultField();
-      }
-      success = authenticator.tryAuthenticate(rfb);
-      if (authenticator.isInteractionNecessary()) {
-	vncContainer.remove(authenticator);
+      if (passwordParam != null) {
+        rfb.authenticateVNC(passwordParam);
       } else {
-	// Don't retry non-interactive authentication.
-	if (!success)
-	  throw new Exception("VNC authentication failed");
+        String pw = askPassword();
+        rfb.authenticateVNC(pw);
       }
       break;
-    case RfbProto.AuthUnixLogin:
-      showConnectionStatus("Performing Unix login-style authentication");
-      showAuthPanel(authenticatorUnixLogin);
-      authenticatorUnixLogin.moveFocusToDefaultField();
-      success = authenticatorUnixLogin.tryAuthenticate(rfb);
-      vncContainer.remove(authenticatorUnixLogin);
-      break;
     default:
       throw new Exception("Unknown authentication scheme " + authType);
     }
-
-    if (!success)
-      rfb.close();
-
-    return success;
   }
 
 
@@ -416,10 +410,12 @@ public class VncViewer extends java.applet.Applet
   // Show an authentication panel.
   //
 
-  void showAuthPanel(Panel authPanel)
+  String askPassword() throws Exception
   {
     showConnectionStatus(null);
 
+    AuthPanel authPanel = new AuthPanel(this);
+
     GridBagConstraints gbc = new GridBagConstraints();
     gbc.gridwidth = GridBagConstraints.REMAINDER;
     gbc.anchor = GridBagConstraints.NORTHWEST;
@@ -435,6 +431,12 @@ public class VncViewer extends java.applet.Applet
     } else {
       validate();
     }
+
+    authPanel.moveFocusToDefaultField();
+    String pw = authPanel.getPassword();
+    vncContainer.remove(authPanel);
+
+    return pw;
   }
 
 
@@ -475,8 +477,8 @@ public class VncViewer extends java.applet.Applet
     if (preferredEncoding == -1) {
       long kbitsPerSecond = rfb.kbitsPerSecond();
       if (nEncodingsSaved < 1) {
-        // Choose Tight encoding for the very first update.
-        System.out.println("Using Tight encoding");
+        // Choose Tight or ZRLE encoding for the very first update.
+        System.out.println("Using Tight/ZRLE encodings");
         preferredEncoding = RfbProto.EncodingTight;
       } else if (kbitsPerSecond > 2000 &&
                  encodingsSaved[0] != RfbProto.EncodingHextile) {
@@ -486,9 +488,9 @@ public class VncViewer extends java.applet.Applet
         preferredEncoding = RfbProto.EncodingHextile;
       } else if (kbitsPerSecond < 1000 &&
                  encodingsSaved[0] != RfbProto.EncodingTight) {
-        // Switch to Tight if the connection speed is below 1Mbps.
+        // Switch to Tight/ZRLE if the connection speed is below 1Mbps.
         System.out.println("Throughput " + kbitsPerSecond +
-                           " kbit/s - changing to Tight encoding");
+                           " kbit/s - changing to Tight/ZRLE encodings");
         preferredEncoding = RfbProto.EncodingTight;
       } else {
         // Don't change the encoder.
@@ -510,12 +512,15 @@ public class VncViewer extends java.applet.Applet
       encodings[nEncodings++] = RfbProto.EncodingCopyRect;
     }
 
-    if (preferredEncoding != RfbProto.EncodingHextile) {
-      encodings[nEncodings++] = RfbProto.EncodingHextile;
-    }
     if (preferredEncoding != RfbProto.EncodingTight) {
       encodings[nEncodings++] = RfbProto.EncodingTight;
     }
+    if (preferredEncoding != RfbProto.EncodingZRLE) {
+      encodings[nEncodings++] = RfbProto.EncodingZRLE;
+    }
+    if (preferredEncoding != RfbProto.EncodingHextile) {
+      encodings[nEncodings++] = RfbProto.EncodingHextile;
+    }
     if (preferredEncoding != RfbProto.EncodingZlib) {
       encodings[nEncodings++] = RfbProto.EncodingZlib;
     }
@@ -679,7 +684,7 @@ public class VncViewer extends java.applet.Applet
   // those expected in the html applet tag source.
   //
 
-  public void readParameters() {
+  void readParameters() {
     host = readParameter("HOST", !inAnApplet);
     if (host == null) {
       host = getCodeBase().getHost();
@@ -688,9 +693,12 @@ public class VncViewer extends java.applet.Applet
       }
     }
 
-    String str = readParameter("PORT", true);
-    port = Integer.parseInt(str);
+    port = readIntParameter("PORT", 5900);
 
+    // Read "ENCPASSWORD" or "PASSWORD" parameter if specified.
+    readPasswordParameters();
+
+    String str;
     if (inAnApplet) {
       str = readParameter("Open New Window", false);
       if (str != null && str.equalsIgnoreCase("Yes"))
@@ -719,12 +727,45 @@ public class VncViewer extends java.applet.Applet
     // Fine tuning options.
     deferScreenUpdates = readIntParameter("Defer screen updates", 20);
     deferCursorUpdates = readIntParameter("Defer cursor updates", 10);
-    deferUpdateRequests = readIntParameter("Defer update requests", 50);
+    deferUpdateRequests = readIntParameter("Defer update requests", 0);
+
+    // Debugging options.
+    debugStatsExcludeUpdates = readIntParameter("DEBUG_XU", 0);
+    debugStatsMeasureUpdates = readIntParameter("DEBUG_CU", 0);
 
     // SocketFactory.
     socketFactory = readParameter("SocketFactory", false);
   }
 
+  //
+  // Read password parameters. If an "ENCPASSWORD" parameter is set,
+  // then decrypt the password into the passwordParam string. Otherwise,
+  // try to read the "PASSWORD" parameter directly to passwordParam.
+  //
+
+  private void readPasswordParameters() {
+    String encPasswordParam = readParameter("ENCPASSWORD", false);
+    if (encPasswordParam == null) {
+      passwordParam = readParameter("PASSWORD", false);
+    } else {
+      // ENCPASSWORD is hexascii-encoded. Decode.
+      byte[] pw = {0, 0, 0, 0, 0, 0, 0, 0};
+      int len = encPasswordParam.length() / 2;
+      if (len > 8)
+        len = 8;
+      for (int i = 0; i < len; i++) {
+        String hex = encPasswordParam.substring(i*2, i*2+2);
+        Integer x = new Integer(Integer.parseInt(hex, 16));
+        pw[i] = x.byteValue();
+      }
+      // Decrypt the password.
+      byte[] key = {23, 82, 107, 6, 35, 78, 88, 7};
+      DesCipher des = new DesCipher(key);
+      des.decrypt(pw, 0, pw, 0);
+      passwordParam = new String(pw);
+    }
+  }
+
   public String readParameter(String name, boolean required) {
     if (inAnApplet) {
       String s = getParameter(name);
@@ -764,19 +805,13 @@ public class VncViewer extends java.applet.Applet
   }
 
   //
-  // moveFocusToDesktop() - move keyboard focus either to the
-  // VncCanvas or to the AuthPanel.
+  // moveFocusToDesktop() - move keyboard focus either to VncCanvas.
   //
 
   void moveFocusToDesktop() {
     if (vncContainer != null) {
-      if (vc != null && vncContainer.isAncestorOf(vc)) {
+      if (vc != null && vncContainer.isAncestorOf(vc))
 	vc.requestFocus();
-      } else if (vncContainer.isAncestorOf(authenticator)) {
-	authenticator.moveFocusToDefaultField();
-      } else if (vncContainer.isAncestorOf(authenticatorUnixLogin)) {
-	authenticatorUnixLogin.moveFocusToDefaultField();
-      }
     }
   }
 
@@ -785,7 +820,37 @@ public class VncViewer extends java.applet.Applet
   //
 
   synchronized public void disconnect() {
-    System.out.println("Disconnect");
+    System.out.println("Disconnecting");
+
+    if (vc != null) {
+      double sec = (System.currentTimeMillis() - vc.statStartTime) / 1000.0;
+      double rate = Math.round(vc.statNumUpdates / sec * 100) / 100.0;
+      int nRealRects = vc.statNumPixelRects;
+      int nPseudoRects = vc.statNumTotalRects - vc.statNumPixelRects;
+      System.out.println("Updates received: " + vc.statNumUpdates + " (" +
+                         nRealRects + " rectangles + " + nPseudoRects +
+                         " pseudo), " + rate + " updates/sec");
+      int numRectsOther = nRealRects - vc.statNumRectsTight
+        - vc.statNumRectsZRLE - vc.statNumRectsHextile
+        - vc.statNumRectsRaw - vc.statNumRectsCopy;
+      System.out.println("Rectangles:" +
+                         " Tight=" + vc.statNumRectsTight +
+                         "(JPEG=" + vc.statNumRectsTightJPEG +
+                         ") ZRLE=" + vc.statNumRectsZRLE +
+                         " Hextile=" + vc.statNumRectsHextile +
+                         " Raw=" + vc.statNumRectsRaw +
+                         " CopyRect=" + vc.statNumRectsCopy +
+                         " other=" + numRectsOther);
+
+      int raw = vc.statNumBytesDecoded;
+      int compressed = vc.statNumBytesEncoded;
+      if (compressed > 0) {
+          double ratio = Math.round((double)raw / compressed * 1000) / 1000.0;
+          System.out.println("Pixel data: " + vc.statNumBytesDecoded +
+                             " bytes, " + vc.statNumBytesEncoded +
+                             " compressed, ratio " + ratio);
+      }
+    }
 
     if (rfb != null && !rfb.closed())
       rfb.close();
@@ -929,22 +994,11 @@ public class VncViewer extends java.applet.Applet
     }
   }
 
-  //
-  // Move the keyboard focus to the password field on window activation.
-  //
-
-  public void windowActivated(WindowEvent evt) {
-    if (vncFrame.isAncestorOf(authenticator)) {
-      authenticator.moveFocusToDefaultField();
-    } else if (vncContainer.isAncestorOf(authenticatorUnixLogin)) {
-      authenticatorUnixLogin.moveFocusToDefaultField();
-    }
-  }
-
   //
   // Ignore window events we're not interested in.
   //
 
+  public void windowActivated(WindowEvent evt) {}
   public void windowDeactivated (WindowEvent evt) {}
   public void windowOpened(WindowEvent evt) {}
   public void windowClosed(WindowEvent evt) {}
